@@ -1,6 +1,7 @@
 __copyright__ = """
 Copyright (C) 2009-2017 Andreas Kloeckner
 Copyright (C) 2014-2017 Aaron Meurer
+Copyright (C) 2024 Gerhard Sittig
 """
 
 __license__ = """
@@ -25,7 +26,10 @@ THE SOFTWARE.
 
 
 import logging
+import sys
 from datetime import datetime
+from enum import Enum, auto
+
 
 logfile = [None]
 
@@ -49,7 +53,7 @@ class TerminalOrStreamHandler(logging.StreamHandler):
     stderr, depending on whether the debugger is active.
     """
     def emit(self, record):
-        from pudb import _have_debugger, _get_debugger
+        from pudb import _get_debugger, _have_debugger
         logfile = getlogfile()
 
         self.acquire()
@@ -94,17 +98,22 @@ ui_log, settings_log = _init_loggers()
 # {{{ breakpoint validity
 
 def generate_executable_lines_for_code(code):
-    lineno = code.co_firstlineno
-    yield lineno
-    # See https://github.com/python/cpython/blob/master/Objects/lnotab_notes.txt
-
-    for line_incr in code.co_lnotab[1::2]:
-        # NB: This code is specific to Python 3.6 and higher
-        # https://github.com/python/cpython/blob/v3.6.0/Objects/lnotab_notes.txt
-        if line_incr >= 0x80:
-            line_incr -= 0x100
-        lineno += line_incr
+    if sys.version_info >= (3, 10):
+        for _start, _end, lineno in code.co_lines():
+            if lineno is not None:
+                yield lineno
+    else:
+        lineno = code.co_firstlineno
         yield lineno
+        # See https://github.com/python/cpython/blob/master/Objects/lnotab_notes.txt
+
+        for line_incr in code.co_lnotab[1::2]:
+            # NB: This code is specific to Python 3.6 and higher
+            # https://github.com/python/cpython/blob/v3.6.0/Objects/lnotab_notes.txt
+            if line_incr >= 0x80:
+                line_incr -= 0x100
+            lineno += line_incr
+            yield lineno
 
 
 def get_executable_lines_for_codes_recursive(codes):
@@ -165,7 +174,7 @@ def lookup_module(filename):
     f = os.path.join(sys.path[0], filename)
     if os.path.exists(f):  # and self.canonic(f) == self.mainpyfile:
         return f
-    root, ext = os.path.splitext(filename)
+    _root, ext = os.path.splitext(filename)
     if ext == "":
         filename = filename + ".py"
     if os.path.isabs(filename):
@@ -185,14 +194,16 @@ def lookup_module(filename):
 # the main idea stolen from Python 3.1's tokenize.py, by Ka-Ping Yee
 
 import re
+
+
 cookie_re = re.compile(br"^\s*#.*coding[:=]\s*([-\w.]+)")
-from codecs import lookup, BOM_UTF8
+from codecs import BOM_UTF8, lookup
 
 
 def detect_encoding(line_iter):
     """
     The detect_encoding() function is used to detect the encoding that should
-    be used to decode a Python source file. It requires one argment, line_iter,
+    be used to decode a Python source file. It requires one argument, line_iter,
     an iterator on the lines to be read.
 
     It will read a maximum of two lines, and return the encoding used
@@ -226,9 +237,9 @@ def detect_encoding(line_iter):
         encoding = matches[0].decode()
         try:
             codec = lookup(encoding)
-        except LookupError:
+        except LookupError as err:
             # This behaviour mimics the Python interpreter
-            raise SyntaxError("unknown encoding: " + encoding)
+            raise SyntaxError("unknown encoding: " + encoding) from err
 
         if bom_found and codec.name != "utf-8":
             # This behaviour mimics the Python interpreter
@@ -274,4 +285,81 @@ def decode_lines(lines):
 
 # }}}
 
+
+# {{{ get single key press from console outside of curses
+
+class KeyReadImpl(Enum):
+    INPUT = auto()
+    GETCH = auto()
+    SELECT = auto()
+
+
+_keyread_impl = KeyReadImpl.INPUT
+if sys.platform in ("emscripten", "wasi"):
+    pass
+elif sys.platform in ("win32",):
+    _keyread_impl = KeyReadImpl.GETCH
+else:
+    _keyread_impl = KeyReadImpl.SELECT
+
+
+class ConsoleSingleKeyReader:
+    """
+    Get a single key press from a terminal without a prompt.
+
+    Eliminates the necessity to press ENTER before other input also
+    becomes available. Avoids the accumulation of prompts on the screen
+    as was the case with Python's input() call. Is used in situations
+    where urwid is disabled and curses calls are not available.
+
+    Supports major desktop platforms with special cases (msvcrt getch(),
+    termios and select). Transparently falls back to Python's input()
+    method. Call sites remain simple and straight forward.
+    """
+
+    def __enter__(self):
+        if _keyread_impl == KeyReadImpl.SELECT:
+            import termios
+            import tty
+            self.prev_settings = termios.tcgetattr(sys.stdin)
+            tty.setcbreak(sys.stdin.fileno())
+        return self
+
+    def __exit__(self, type, value, traceback):
+        if _keyread_impl == KeyReadImpl.SELECT:
+            import termios
+            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.prev_settings)
+
+    def get_single_key(self):
+        if _keyread_impl == KeyReadImpl.GETCH:
+            import msvcrt
+            # https://docs.python.org/3/library/msvcrt.html#msvcrt.getch
+            # Most keys are returned in the first getch() call. Some
+            # special keys (function keys, cursor, keypad) require
+            # another call when the first returned '\0' or '\xe0'.
+            c = msvcrt.getch()
+            if c in ("\x00", "\xe0"):
+                c = msvcrt.getch()
+            return c
+
+        elif _keyread_impl == KeyReadImpl.SELECT:
+            import select
+            rset, _, _ = select.select([sys.stdin], [], [], None)
+            assert sys.stdin in rset
+            return sys.stdin.read(1)
+
+        # Strictly speaking putting the fallback here which requires
+        # pressing ENTER is not correct, this is the "non buffered"
+        # console support code. But it simplifies call sites. And is
+        # easy to tell by users because a prompt is provided. This is
+        # the most portable approach, and backwards compatible with
+        # earlier PuDB releases. It's a most appropriate default for
+        # otherwise unsupported platforms. Or when users choose to
+        # not accept single key presses, or keys other than ENTER.
+        else:
+            input("Hit Enter to return:")
+            return None
+
+# }}}
+
 # vim: foldmethod=marker
