@@ -25,20 +25,17 @@ THE SOFTWARE.
 
 import os
 import sys
-
 from configparser import ConfigParser
-from pudb.lowlevel import (lookup_module, get_breakpoint_invalid_reason,
-                           settings_log)
 
-# minor LGPL violation: stolen from python-xdg
-
-_home = os.environ.get("HOME", None)
-xdg_data_home = os.environ.get("XDG_DATA_HOME",
-            os.path.join(_home, ".local", "share") if _home else None)
+from pudb.lowlevel import get_breakpoint_invalid_reason, lookup_module, settings_log
 
 
-XDG_CONFIG_HOME = os.environ.get("XDG_CONFIG_HOME",
-                                 os.path.join(_home, ".config") if _home else None)
+# see https://github.com/inducer/pudb/pull/453 for context
+_home = os.environ.get("HOME", os.path.expanduser("~"))
+XDG_CONF_RESOURCE = "pudb"
+XDG_CONFIG_HOME = os.environ.get(
+    "XDG_CONFIG_HOME",
+    os.path.join(_home, ".config") if _home else None)
 
 if XDG_CONFIG_HOME:
     XDG_CONFIG_DIRS = [XDG_CONFIG_HOME]
@@ -46,26 +43,21 @@ else:
     XDG_CONFIG_DIRS = os.environ.get("XDG_CONFIG_DIRS", "/etc/xdg").split(":")
 
 
-def get_save_config_path(*resource):
-    if XDG_CONFIG_HOME is None:
+def get_save_config_path():
+    # This may not raise, as it is called during import.
+    if not XDG_CONFIG_HOME:
         return None
-    if not resource:
-        resource = [XDG_CONF_RESOURCE]
 
-    # no idea what pylint's problem is here
-    resource = os.path.join(*resource)  # pylint: disable=no-value-for-parameter
+    path = os.path.join(XDG_CONFIG_HOME, XDG_CONF_RESOURCE)
+    try:
+        os.makedirs(path, mode=0o700, exist_ok=True)
+    except Exception:
+        settings_log.exception("Failed to make config dir")
 
-    assert not resource.startswith("/")
-    path = os.path.join(XDG_CONFIG_HOME, resource)
-    if not os.path.isdir(path):
-        os.makedirs(path, 448)  # 0o700
     return path
 
-# end LGPL violation
-
 
 CONF_SECTION = "pudb"
-XDG_CONF_RESOURCE = "pudb"
 CONF_FILE_NAME = "pudb.cfg"
 
 SAVED_BREAKPOINTS_FILE_NAME = "saved-breakpoints-%d.%d" % sys.version_info[:2]
@@ -76,11 +68,13 @@ _config_ = [None]
 
 
 def load_config():
+    # This may not raise, as it is called during import.
+
     # Only ever do this once
     if _config_[0] is not None:
         return _config_[0]
 
-    from os.path import join, isdir
+    from os.path import isdir, join
 
     cparser = ConfigParser()
 
@@ -122,6 +116,13 @@ def load_config():
 
     conf_dict.setdefault("hide_cmdline_win", "False")
 
+    # hotkeys
+    conf_dict.setdefault("hotkeys_code", "C")
+    conf_dict.setdefault("hotkeys_variables", "V")
+    conf_dict.setdefault("hotkeys_stack", "S")
+    conf_dict.setdefault("hotkeys_breakpoints", "B")
+    conf_dict.setdefault("hotkeys_toggle_cmdline_focus", "ctrl x")
+
     def normalize_bool_inplace(name):
         try:
             if conf_dict[name].lower() in ["0", "false", "off"]:
@@ -141,6 +142,8 @@ def load_config():
 
 
 def save_config(conf_dict):
+    # This may not raise, as it is called during import.
+
     from os.path import join
 
     cparser = ConfigParser()
@@ -264,7 +267,8 @@ def edit_config(ui, conf_dict):
             bool(conf_dict["prompt_on_quit"]), on_state_change=_update_config,
                 user_data=("prompt_on_quit", None))
 
-    hide_cmdline_win = urwid.CheckBox("Hide command line (Ctrl-X) window "
+    hide_cmdline_win = urwid.CheckBox("Hide command line"
+            f"({conf_dict['hotkeys_toggle_cmdline_focus']}) window "
                                       "when not in use",
             bool(conf_dict["hide_cmdline_win"]), on_state_change=_update_config,
                 user_data=("hide_cmdline_win", None))
@@ -277,7 +281,7 @@ def edit_config(ui, conf_dict):
               "ptpython", "ptipython"]
     known_shell = conf_dict["shell"] in shells
     shell_edit = urwid.Edit(edit_text=conf_dict["custom_shell"])
-    shell_edit_list_item = urwid.AttrMap(shell_edit, "value")
+    shell_edit_list_item = urwid.AttrMap(shell_edit, "input", "focused input")
 
     shell_rb_group = []
     shell_rbs = [
@@ -288,7 +292,7 @@ def edit_config(ui, conf_dict):
                 not known_shell, on_state_change=_update_config,
                 user_data=("shell", None)),
                 shell_edit_list_item,
-                urwid.Text("\nTo use a custom shell, see example-shell.py "
+                urwid.Text("\nTo use a custom shell, see examples/shell.py "
                     "in the pudb distribution. Enter the full path to a "
                     "file like it in the box above. '~' will be expanded "
                     "to your home directory. The file should contain a "
@@ -301,13 +305,13 @@ def edit_config(ui, conf_dict):
 
     # {{{ themes
 
-    from pudb.theme import THEMES
+    from pudb.themes import THEMES
 
     known_theme = conf_dict["theme"] in THEMES
 
     theme_rb_group = []
     theme_edit = urwid.Edit(edit_text=conf_dict["custom_theme"])
-    theme_edit_list_item = urwid.AttrMap(theme_edit, "value")
+    theme_edit_list_item = urwid.AttrMap(theme_edit, "input", "focused input")
     theme_rbs = [
             urwid.RadioButton(theme_rb_group, name,
                 conf_dict["theme"] == name, on_state_change=_update_config,
@@ -317,7 +321,7 @@ def edit_config(ui, conf_dict):
                     not known_theme, on_state_change=_update_config,
                     user_data=("theme", None)),
                 theme_edit_list_item,
-                urwid.Text("\nTo use a custom theme, see example-theme.py in the "
+                urwid.Text("\nTo use a custom theme, see examples/theme.py in the "
                     "pudb distribution. Enter the full path to a file like it in "
                     "the box above. '~' will be expanded to your home directory. "
                     "Note that a custom theme will not be applied until you close "
@@ -354,7 +358,8 @@ def edit_config(ui, conf_dict):
         "selecting a variable and typing 'e' to edit the variable's display "
         "settings, or by typing one of d/t/r/s/i/c. Note that str and repr will "
         "be slower than the default, type, or id stringifiers.\n")
-    stringifier_edit_list_item = urwid.AttrMap(stringifier_edit, "value")
+    stringifier_edit_list_item = urwid.AttrMap(stringifier_edit,
+                                               "input", "focused input")
     stringifier_rbs = [
             urwid.RadioButton(stringifier_rb_group, name,
                 conf_dict["stringifier"] == name,
@@ -367,7 +372,7 @@ def edit_config(ui, conf_dict):
                     user_data=("stringifier", None)),
                 stringifier_edit_list_item,
                 urwid.Text("\nTo use a custom stringifier, see "
-                    "example-stringifier.py in the pudb distribution. Enter the "
+                    "examples/stringifier.py in the pudb distribution. Enter the "
                     "full path to a file like it in the box above. "
                     "'~' will be expanded to your home directory. "
                     "The file should contain a function called pudb_stringifier() "
@@ -572,7 +577,7 @@ def load_breakpoints():
     Loads and check saved breakpoints out from files
     Returns: list of tuples
     """
-    from os.path import join, isdir
+    from os.path import isdir, join
 
     file_names = []
     for cdir in XDG_CONFIG_DIRS:
